"""
方便开发者配置设置的工具。
"""
from __future__ import annotations

__all__ = [
    'dict_',
    'datasize',
    'during',
]

import re
from itertools import chain
from typing import Any


def dict_(*pairs: tuple[str, Any], **kwargs: Any) -> dict:
    """
    :class:`dict` 的变种。

    - 会去除“键”尾随的所有 ``_`` 。
    - 接收任意个二元元组（元组形式的键-值对）。
    - 注意，此函数不能完全兼容 :class:`dict` 方法。

    >>> dict_(
    >>>     ('.', {'foo': 'baz'}),
    >>>     ('level', 'DEBUG'),
    >>>     class_='logging.StreamHandler',
    >>>     filters=[],
    >>>     formatter='bear',
    >>> )
    {
        '.': {
            'foo': 'baz'
        },
        'level': 'DEBUG',
        'class': 'logging.StreamHandler',
        'filters': [],
        'formatter': 'bear',
    }
    """
    return dict(chain(
        pairs,
        ((k.rstrip('_'), v) for k, v in kwargs.items()),
    ))


class datasize:
    PATTERN = re.compile(r'^([0-9]+)\s*([KMGTPEZY]?)(i?[Bb])$')
    SEPERATOR = ','

    def __new__(cls, literal: str) -> int | float:
        """
        数据大小字面值 → 字节数目。

        - 字面值由一个或多个部分组成，使用英文逗号 ``,`` 分隔，每个部分是整数+单位，字节数目由所有部分累加求和得到。
          只要不涉及比特 ``b`` ，那么返回结果必定为 :class:`int` 类型，否则必定为 :class:`float` 类型。
        - 整数部分不支持小数点、仅支持十进制、不支持负号。
        - 单位部分严格区分大小写。
        - 任一部分解析失败都会抛出 :class:`ValueError` 以避免潜在的错误。
        - 强烈建议在单位及每个部分之间预留空格，方便开发和运维人员检查！！！

        >>> datasize('10KB')
        10000
        >>> datasize('10KiB')
        10240
        >>> datasize('10 KiB, 1 B')
        10241
        >>> datasize('10 KiB, 1 MiB')
        1058816
        >>> datasize('10 KiB, 1 b')
        10240.125

        支持的单位包括：（mB 这类并非正式缩写，故不支持）

        - B、b
        - KB、KiB、Kb、Kib
        - MB、MiB、Mb、Mib
        - GB、GiB、Gb、Gib
        - TB、TiB、Tb、Tib
        - 以此类推……

        它们之间的进位转换关系如下：

        - 1 B == 8 b
        - 1 MB == 1000 KB
        - 1 MiB == 1024 KiB
        """
        if not isinstance(literal, str):
            raise TypeError(
                'cannot parse an non-string value.'
            )
        return sum(cls._parse(part) for part in literal.split(cls.SEPERATOR))

    @classmethod
    def _parse(cls, literal: str) -> int | float:
        result = re.fullmatch(cls.PATTERN, literal)
        if result is None:
            raise ValueError(
                f'cannot parse data-size literal "{literal}", '
                f'plz check and fix it on your configurations.'
            )
        base = int(result.group(1))
        shift = 'BKMGTPEZY'.index(result.group(2))
        power = (1024 if 'i' in result.group(3) else 1000) ** shift
        power = (power / 8) if 'b' in result.group(3) else power
        return base * power


class during:
    UNITS = dict(s=1, m=60, h=3600, d=86400, w=86400 * 7)
    PATTERN = re.compile(r'^([0-9]+)\s*([wdhms])$')
    SEPERATOR = ','

    def __new__(cls, literal: str) -> int:
        """
        时长字面值 → 秒数。

        - 字面值由一个或多个部分组成，使用英文逗号 ``,`` 分隔，每个部分是整数+单位，字节数目由所有部分累加求和得到。
        - 整数部分不支持小数点、仅支持十进制、不支持负号。
        - 单位部分严格区分大小写。
        - 任一部分解析失败都会抛出 :class:`ValueError` 以避免潜在的错误。
        - 强烈建议在单位及每个部分之间预留空格，方便开发和运维人员检查！！！

        >>> during('1h')
        3600
        >>> during('1d')
        86400
        >>> during('1h,1m,1s')
        3661

        支持的单位及之间的转换关系如下：

        - 1w == 7d
        - 1d == 24h == 86400s
        - 1h == 60m == 3600s
        - 1m == 60s
        """
        if not isinstance(literal, str):
            raise TypeError(
                'cannot parse an non-string value.'
            )
        return sum(cls._parse(part) for part in literal.split(cls.SEPERATOR))

    @classmethod
    def _parse(cls, literal: str) -> int | float:
        result = re.fullmatch(cls.PATTERN, literal)
        if result is None:
            raise ValueError(
                f'cannot parse duration literal "{literal}", '
                f'plz check and fix it on your configurations.'
            )
        base = int(result.group(1))
        power = cls.UNITS[result.group(2)]
        return base * power



"""
字符、字符串相关工具与常量。
"""
__all__ = [
    'Notations',
    'Chars',
    'SafeChars',
    'randb64',
    'randb64y',
    'randb62',
    'randb16',
    'case_camel_to_snake',
]

import os
import re
from base64 import b64encode, urlsafe_b64encode
from itertools import chain
from math import ceil
from random import getrandbits


class Notations:
    """
    进位制数码。
    """
    BASE8 = '01234567'
    """包含：阿拉伯数字 ``0`` 到 ``7`` 。"""
    BASE10 = '0123456789'
    """包含：阿拉伯数字 ``0`` 到 ``9`` 。"""
    BASE16 = ''.join(chain(BASE10, map(chr, range(ord('A'), ord('F') + 1))))
    """包含：阿拉伯数字 ``0`` 到 ``9``、大写字母 ``A`` 到 ``F`` 。"""
    BASE36 = ''.join(chain(BASE10, map(chr, range(ord('A'), ord('Z') + 1))))
    """包含：阿拉伯数字 ``0`` 到 ``9``、大写字母 ``A`` 到 ``Z`` 。"""
    BASE62 = ''.join(chain(BASE36, map(chr, range(ord('a'), ord('z') + 1))))
    """包含：阿拉伯数字 ``0`` 到 ``9``、大写字母 ``A`` 到 ``Z``、小写字母 ``a`` 到 ``z``。"""
    BASE64 = BASE62 + '+/'
    """包含：阿拉伯数字 ``0`` 到 ``9``、大写字母 ``A`` 到 ``Z``、小写字母 ``a`` 到 ``z``、符号 ``+`` 和 ``/``。"""
    BASE64SAFE = BASE62 + '-_'
    """包含：阿拉伯数字 ``0`` 到 ``9``、大写字母 ``A`` 到 ``Z``、小写字母 ``a`` 到 ``z``、符号 ``-`` 和 ``_``。"""


class Chars:
    """
    不同类型的字符。
    """
    DIGIT = Notations.BASE10
    """包含：阿拉伯数字 ``0`` 到 ``9`` 。"""
    UPPER = ''.join(map(chr, range(ord('A'), ord('Z') + 1)))
    """包含：大写字母 ``A`` 到 ``Z`` 。"""
    LOWER = ''.join(map(chr, range(ord('a'), ord('z') + 1)))
    """包含：小写字母 ``a`` 到 ``z``。"""
    LETTER = UPPER + LOWER
    """包含：大写字母 ``A`` 到 ``Z``、小写字母 ``a`` 到 ``z``。"""
    SYMBOL = ''.join(chain(
        map(chr, range(0x21, 0x2F + 1)),
        map(chr, range(0x3A, 0x40 + 1)),
        map(chr, range(0x5B, 0x60 + 1)),
        map(chr, range(0x7B, 0x7E + 1)),
    ))
    """包含 ``0x21`` 到 ``0x7e`` 之间的所有可见符号，即除了空格、控制符、数字、大小写字母。"""


class SafeChars:
    DIGIT = Chars.DIGIT.replace('0', '').replace('1', '')
    """除了 ``0`` 和 ``1`` 以外的阿拉伯数字。"""
    UPPER = Chars.UPPER.replace('I', '').replace('O', '')
    """包含：大写字母 ``A`` 到 ``Z``，除了 ``I`` 和 ``O``。"""
    LOWER = Chars.LOWER.replace('l', '')
    """包含：小写字母 ``a`` 到 ``z``，除了 ``l``。"""
    LETTER = UPPER + LOWER
    """包含：大写字母 ``A`` 到 ``Z`` 和小写字母 ``a`` 到 ``z``，除了大写字母 ``I``、``O`` 和小写字母 ``l``。"""
    SYMBOL_NORMAL = r"`-=[]\;',./"
    SYMBOL_SHIFT = r'~!@#$%^&*()_+{}|:"<>?'


def randb64(n: int, safe=False, use_os=False) -> str:
    """
    快速生成 n 个 Base64（不包含 ``=``）随机字符。

    - ``safe`` 决定是否生成 URL-Safe Base64 字符串。
    - ``use_os=False`` 时会受 random.seed() 影响。
    - ``use_os=True`` 则使用 :mod:`os` 库，在大量调用时可能会耗费略少的时间。
    """
    if n < 1:
        return ''
    if not safe:
        if use_os:
            return b64encode(os.urandom(ceil(n * 6 / 8))).decode('ASCII')[:n]
        else:
            return b64encode(getrandbits(n * 6).to_bytes(ceil(n * 6 / 8), 'little')).decode('ASCII')[:n]
    else:
        if use_os:
            return urlsafe_b64encode(os.urandom(ceil(n * 6 / 8))).decode('ASCII')[:n]
        else:
            return urlsafe_b64encode(getrandbits(n * 6).to_bytes(ceil(n * 6 / 8), 'little')).decode('ASCII')[:n]


def randb64y(n: int, safe=False, use_os=False) -> str:
    """
    快速生成 n 字节的 Base64 随机字符（可能包含 ``=``）。

    - 不移除尾缀的 ``=`` 则生成长度为 ``ceil(n/3)*4`` 。
    - 移除尾缀的 ``=`` 的话生成长度为 ``ceil(n/3*4)`` 。
    - ``safe`` 决定是否生成 URL-Safe Base64 字符串。
    - ``use_os=False`` 时会受 random.seed() 影响。
    - ``use_os=True`` 则使用 :mod:`os` 库，在大量调用时可能会耗费略少的时间。

    >>> randb64y(13)
    NP4W8LAhbqz6sSRuNg==

    >>> randb64y(13).rstrip('=')
    NP4W8LAhbqz6sSRuNg
    """
    if n < 1:
        return ''
    if not safe:
        if use_os:
            return b64encode(os.urandom(n)).decode('ASCII')
        else:
            return b64encode(getrandbits(n * 8).to_bytes(n, 'little')).decode('ASCII')
    else:
        if use_os:
            return urlsafe_b64encode(os.urandom(n)).decode('ASCII')
        else:
            return urlsafe_b64encode(getrandbits(n * 8).to_bytes(n, 'little')).decode('ASCII')


def randb62(n: int, use_os=False) -> str:
    """
    生成 n 个 Base62 随机字符。

    - ``use_os=False`` 时会受 random.seed() 影响。
    - ``use_os=True`` 则使用 :mod:`os` 库，在大量调用时可能会耗费略少的时间。
    """
    if n < 1:
        return ''
    if use_os:
        return ''.join(Notations.BASE62[i % 62] for i in os.urandom(n))
    else:
        return ''.join(Notations.BASE62[i % 62] for i in getrandbits(n * 8).to_bytes(n, 'little'))


def randb16(n: int, use_os=False) -> str:
    """
    生成 n 个 Base16（即 Hex）随机字符。

    - ``use_os=False`` 时会受 random.seed() 影响。
    - ``use_os=True`` 则使用 :mod:`os` 库，在大量调用时可能会耗费略少的时间。

    >>> randb16(8)
    'd7d3d2ed'

    >>> randb16(8).upper()
    'D7D3D2ED'
    """
    if n < 1:
        return ''
    if use_os:
        return os.urandom(ceil(n / 2)).hex()[:n]
    else:
        return getrandbits(n * 4).to_bytes(ceil(n / 2), 'little').hex()[:n]


def case_camel_to_snake(name: str) -> str:
    """
    将类似 ``CombineOrderSKUModel`` 大小写形式的字符串
    转换为 ``combine_order_sku_model`` 。
    """
    # "CombineOrderSKUModel"
    # -> "Combine OrderSKU Model"
    # -> "Combine Order SKU Model"
    # -> "combine_order_sku_model"
    mid = re.sub('(.)([A-Z][a-z]+)', r'\1 \2', name)
    words = re.sub('([a-z0-9])([A-Z])', r'\1 \2', mid)
    return '_'.join(word.lower() for word in words.split())



"""
对 `Django REST Framework <https://www.django-rest-framework.org/>`_ 的扩展和增强。
"""
from __future__ import annotations

__all__ = [
    'BearerAuthentication',
    'EasyViewSetMixin',
    'SoftDeleteModelMixin',
    'ExistingFilterBackend',
    'ActiveStatusFilterBackend',
]

from typing import Any

from django.core.exceptions import FieldDoesNotExist
from django.utils.decorators import classonlymethod
from rest_framework import status
from rest_framework.authentication import TokenAuthentication
from rest_framework.authtoken.models import Token
from rest_framework.filters import BaseFilterBackend
from rest_framework.response import Response
from rest_framework.viewsets import ViewSetMixin


class BearerAuthentication(TokenAuthentication):
    """
    Simple token based authentication.

    Clients should authenticate by passing the token key in the "Authorization"
    HTTP header, prepended with the string "Bearer ".  For example:

        Authorization: Bearer 401f7ac837da42b97f613d789819ff93537bee6a
    """
    keyword = 'Bearer'
    model = Token


class EasyViewSetMixin(ViewSetMixin):
    """
    提供了两个方法来简化 :func:`as_view() <rest_framework.viewsets.ViewSetMixin.as_view>` 的传参。

    适用于：

    - :class:`ViewSet <rest_framework.viewsets.ViewSet>` 的子类
    - :class:`GenericViewSet <rest_framework.viewsets.GenericViewSet>` 的子类
    - 需要 :class:`ViewSetMixin <rest_framework.viewsets.ViewSetMixin>` 的类
    """

    @classonlymethod
    def to_view(cls, initkwargs: dict[str, Any] = None, **actions: str):
        """
        将视图类存储在一个视图函数上，方便URL反向查找。

        之所以互换 *initkwargs* 和 *actions* 的传参方式，是因为后者往往用得更频繁。

        :param initkwargs: 视图类的初始化参数。
        :param actions: HTTP方法及其在视图类中的方法函数名称。
                        不同方法（的组合）可以让不同的视图函数应对不同的HTTP请求。
                        至少要提供一个key-value。
        :return: 一个新的视图函数。
        :raise TypeError: 来自 :func:`as_view() <rest_framework.viewsets.ViewSetMixin.as_view>` 的异常。
        """
        if initkwargs is None:
            initkwargs = {}
        return cls.as_view(actions, **initkwargs)

    @classonlymethod
    def av(cls, actions: str, **initkwargs):
        """
        将视图类存储在一个视图函数上，方便URL反向查找。

        使用区分大小写的字母代表指定视图类方法及其对应的HTTP方法。

        ``actions`` 的缩写如下：
          - ``o``，对应 OPTIONS 请求及 ``.options()`` 方法。
          - ``l``，对应 GET 请求及 ``.list()`` 方法。
          - ``r``，对应 GET 请求及 ``.retrieve()`` 方法。
          - ``c``，对应 POST 请求及 ``.create()`` 方法。
          - ``u``，对应 PUT 请求及 ``.update()`` 方法。
          - ``U``，对应 PUT 请求及 ``.partial_update()`` 方法。
          - ``p``，对应 PATCH 请求及 ``.partial_update()`` 方法。
          - ``d``，对应 DELETE 请求及 ``.destroy()`` 方法。
          - ``D``，对应 DELETE 请求及 ``.soft_delete()`` 方法。

        :param actions: 缩写。相同HTTP请求的缩写不能组合，否则只择其一。
        :param initkwargs: 视图类的初始化参数。
        :return: 一个新的视图函数。
        :raise KeyError: 缩写未被定义。
        :raise TypeError: 来自 :func:`as_view() <rest_framework.viewsets.ViewSetMixin.as_view>` 的异常。
        """
        # 缩写 p 是请求方法的性质向功能的妥协。
        # PATCH 请求是对资源进行部分修改，PUT 请求是对资源进行整体覆盖，与 partial 相悖；
        # 但是 PUT 请求是幂等的，而 PATCH 不是，也就是说多次
        # PATCH 相同内容产生的效果可以不一样，这与 update 相悖。
        mapper = {
            'o': ('options', 'options'),
            'l': ('get', 'list'),
            'r': ('get', 'retrieve'),
            'c': ('post', 'create'),
            'u': ('put', 'update'),
            'U': ('put', 'partial_update'),
            'p': ('patch', 'partial_update'),
            'd': ('delete', 'destroy'),
            'D': ('delete', 'soft_delete'),
        }
        return cls.as_view(dict(mapper[a] for a in actions), **initkwargs)


class SoftDeleteModelMixin:
    """
    将一个模型实例标记为已删除。（软删除）

    - 通过 ``self.deletion_field`` 配置存储标记的字段，默认是 ``deleted``。
    - 通过 ``self.deletion_mark`` 配置标记是什么，默认是布尔值 ``True`` 。

    适用于 :class:`rest_framework.generics.GenericAPIView` 的子类
    """
    deletion_field = 'deleted'
    deletion_mark = True

    # noinspection PyUnresolvedReferences
    def soft_delete(self, request, *args, **kwargs):
        instance = self.get_object()
        self.perform_soft_delete(instance)
        return Response(status=status.HTTP_204_NO_CONTENT)

    def perform_soft_delete(self, instance):
        if not hasattr(instance, self.deletion_field):
            raise TypeError(
                '模型 %s 没有用于标记删除的字段 %s 。' % (
                    type(instance).__name__,
                    self.deletion_field,
                )
            )

        setattr(instance, self.deletion_field, self.deletion_mark)

        instance.save()


class ExistingFilterBackend(BaseFilterBackend):
    """
    筛选掉查询集中标记为已删除的结果。

    - 删除标记默认为 ``True`` ，可以通过在视图类中添加 ``deletion_mark`` 属性来更改。
    - 标记删除的字段默认是 ``deleted`` ，可以通过在视图类中添加 ``deletion_field`` 属性来更改。
      若字段不存在，会抛出 :class:`django.core.exceptions.FieldDoesNotExist` 。

    适用于：

    - :class:`rest_framework.viewsets.GenericViewSet` 子类的 ``filter_backends`` 属性
    - ``rest_framework.settings.DEFAULT_FILTER_BACKENDS``
    - ``django.conf.settings.REST_FRAMEWORK["DEFAULT_FILTER_BACKENDS"]``
    """

    def filter_queryset(self, request, queryset, view):
        field = getattr(view, 'deletion_field', 'deleted')
        mark = getattr(view, 'deletion_mark', True)

        if field not in queryset.model.fields:
            raise FieldDoesNotExist(
                '模型 %s 缺少用于标记删除的字段 %s' % (
                    queryset.model._meta.object_name,
                    field,
                ),
            )

        return queryset.exclude(**{field: mark})


class ActiveStatusFilterBackend(BaseFilterBackend):
    """
    筛选查询集中已启用的结果。

    - 启用标记默认为 ``True`` ，可以通过在视图类中添加 ``active_mark`` 属性来更改。
    - 标记启用的字段默认是 ``activated`` ，可以通过在视图类中添加 ``active_field`` 属性来更改。
      若字段不存在，会抛出 :class:`django.core.exceptions.FieldDoesNotExist` 。

    适用于：
      - :class:`rest_framework.viewsets.GenericViewSet` 子类的 ``filter_backends`` 属性
      - ``rest_framework.settings.DEFAULT_FILTER_BACKENDS``
      - ``django.conf.settings.REST_FRAMEWORK["DEFAULT_FILTER_BACKENDS"]``
    """

    def filter_queryset(self, request, queryset, view):
        field = getattr(view, 'active_field', 'activated')
        mark = getattr(view, 'active_mark', True)

        if field not in queryset.model.fields:
            raise FieldDoesNotExist(
                '模型 %s 缺少用于标记删除的字段 %s' % (
                    queryset.model._meta.object_name,
                    field,
                ),
            )

        return queryset.filter(**{field: mark})



"""
对 `Django <https://docs.djangoproject.com/zh-hans/4.2/>`_ 的扩展和增强。
"""
from __future__ import annotations

__all__ = [
    'SnakeModel',
    'HasBits',
    'HasAllBits',
    'NotAnyBits',
]

from typing import Any

from django.apps import apps
from django.db import models

from zeraora.string import case_camel_to_snake


class SnakeModel(models.base.ModelBase):
    """
    为模型生成一个下划线分隔的小写的数据表名（即蛇形命名法）。

    - 通过 ``Meta`` 手动指定的表名不会被覆盖。
    - 只会给非抽象模型（``Meta.abstract == False``）生成表名。

    ----

    用法如下： ::

        # ./apps/mall/models.py

        from django.db import models
        from zeraora.django import SnakeModel

        class GoodsSKUInfo(models.Model, metaclass=SnakeModel):
            name = models.CharField(max_length=64)
            price = models.IntegerField()
            stock = models.IntegerField()

            class Meta:
                abstract = False
                # db_table = ""

    以上模型没有设置 ``Meta.db_table``，
    那么 Django 默认生成的是 ``mall_goodsskuinfo``，
    而使用 SnakeModel 之后会默认生成 ``mall_goods_sku_info`` 。
    """

    def __new__(cls, name, bases, attrs, **kwargs):
        module = attrs.get('__module__', None)
        app_config = apps.get_containing_app_config(module)
        if app_config is None:
            return super().__new__(cls, name, bases, attrs, **kwargs)

        app_name = app_config.label
        model_name = case_camel_to_snake(app_name)
        table_name = f'{app_name}_{model_name}'

        if 'Meta' not in attrs:
            attrs['Meta'] = type('Meta', (), dict(db_table=table_name))

        abstract = getattr(attrs["Meta"], 'abstract', False)
        if not hasattr(attrs["Meta"], 'db_table') and not abstract:
            setattr(attrs['Meta'], 'db_table', table_name)

        return super().__new__(cls, name, bases, attrs, **kwargs)


class PrefilterManager(models.Manager):
    """
    预设过滤的数据管理器。
    """

    def __init__(self, **conditions: Any):
        super().__init__()
        self._conditions = conditions

    def get_queryset(self):
        return super().get_queryset().filter(**self._conditions)


class HasBits(models.Lookup):
    """
    目标字段是否含有指定的任意一个或多个比特。
    """
    lookup_name = 'has_bits'

    def as_sql(self, compiler, connection):
        lhs, lhs_params = self.process_lhs(compiler, connection)
        rhs, rhs_params = self.process_rhs(compiler, connection)
        params = lhs_params + rhs_params
        return '%s & %s != 0' % (lhs, rhs), params


class HasAllBits(models.Lookup):
    """
    目标字段是否含有指定的所有比特。
    """
    lookup_name = 'has_all_bits'

    def as_sql(self, compiler, connection):
        lhs, lhs_params = self.process_lhs(compiler, connection)
        rhs, rhs_params = self.process_rhs(compiler, connection)
        params = lhs_params + rhs_params + rhs_params
        return '%s & %s = %s' % (lhs, rhs, rhs), params


class NotAnyBits(models.Lookup):
    """
    目标字段是否不含给定的任意比特。
    """
    lookup_name = 'not_any_bits'

    def as_sql(self, compiler, connection):
        lhs, lhs_params = self.process_lhs(compiler, connection)
        rhs, rhs_params = self.process_rhs(compiler, connection)
        params = lhs_params + rhs_params
        return '%s & %s = 0' % (lhs, rhs), params



"""
枚举相关工具和类型。
"""
from __future__ import annotations

__all__ = [
    'ItemsMeta',
    'Items',
]

import enum
from typing import Any


class ItemsMeta(enum.EnumMeta):
    """
    一个元类，用于创建带有任意属性的枚举的类。
    """

    def __new__(metacls, classname, bases, classdict, **kwds):
        # 获取属性名（pks）
        pks = classdict.get('__properties__', ())
        if isinstance(pks, str):
            pks = (pks,)
        for pk in pks:
            if not isinstance(pk, str) or pk.startswith('_'):
                raise ValueError(
                    f'{classname}.__properties__ 包含的值必须是字符串且不以下划线 “_” 开头。'
                )
            # https://docs.python.org/zh-cn/3/library/enum.html#supported-sunder-names
            if pk in ('name', 'value'):
                raise AttributeError(
                    f'不必也不能在 {classname}.__properties__ 中定义 name 和 value，'
                    f'它们是原生枚举就已经支持的属性。'
                )
            if pk in ('generate_next_value',):
                raise KeyError(
                    f'不能在 {classname}.__properties__ 中定义 {pk}，'
                    f'因为它会被转化成 _{pk}_ ，而这属于保留名称。'
                )
        pks = tuple(f'_{pk}_' for pk in pks)
        qty = len(pks) + 1  # 等号右侧所有元素的总数

        # 获取属性值（pvs）和枚举值（value）
        pvs_list = []
        for key in classdict._member_names:
            value = classdict[key]
            pvs = ()
            if isinstance(value, tuple) and len(value) == qty:
                value, *pvs = value
                pvs = tuple(pvs)
            pvs_list.append(pvs)
            dict.__setitem__(classdict, key, value)

        # 为每个枚举添加私有属性
        cls = super().__new__(metacls, classname, bases, classdict, **kwds)
        for member, pvs in zip(cls.__members__.values(), pvs_list):
            member.__dict__.update(zip(pks, pvs))

        return enum.unique(cls)

    def __contains__(cls, member):
        if not isinstance(member, enum.Enum):
            # Allow non-enums to match against member values.
            return any(x.value == member for x in cls)
        return super().__contains__(member)

    def __getattr__(cls, name):
        if not isinstance(name, str):
            raise TypeError  # pragma: no cover
        if name[:-1] in cls.__properties__ and name.endswith('s'):
            return [getattr(member, name[:-1]) for member in cls]
        if name[:-2] in cls.__properties__ and name.endswith('es'):
            return [getattr(member, name[:-2]) for member in cls]
        return object.__getattribute__(cls, name)

    # 对 __empty__ 属性的支持是为了与 Django 的 Choices 相兼容，可参见：
    # https://docs.djangoproject.com/zh-hans/4.2/ref/models/fields/#enumeration-types

    @property
    def names(cls) -> list[str]:
        """
        所有枚举成员的名称（定义枚举成员时的全大写变量名）。
        """
        empty = ["__empty__"] if hasattr(cls, "__empty__") else []
        return empty + [member.name for member in cls]

    @property
    def values(cls) -> list:
        """
        所有枚举成员的值（定义枚举成员时等号右边元组的第一个值）。
        """
        empty = [None] if hasattr(cls, "__empty__") else []
        return empty + [member.value for member in cls]

    @property
    def items(cls) -> dict[str, Any]:
        """
        所有枚举成员的名称和值。
        """
        its = {"__empty__": None} if hasattr(cls, "__empty__") else {}
        its.update((member.name, member.value) for member in cls)
        return its

    @property
    def choices(cls) -> list[tuple[str, Any] | tuple[None, Any]]:
        """
        所有枚举成员的值，和所有枚举成员的属性中的标签（label）。
        """
        if 'label' not in cls.__properties__:
            raise AttributeError(
                '使用 .choices 属性前必须在 __properties__ 中'
                '添加一个名为 "label" 的属性，且必须保证枚举值中有相应的属性值。'
            )
        empty = [(None, cls.__empty__)] if hasattr(cls, "__empty__") else []
        return empty + [(member.value, member.label) for member in cls]

    def asdict(cls) -> dict[enum.Enum, Any]:
        """
        返回枚举成员与枚举值之间的映射。
        """
        return {member: member.value for member in cls}


class Items(enum.Enum, metaclass=ItemsMeta):
    """
    用于创建带有任意属性的枚举。

    >>> from enum import Enum
    >>>
    >>> class Grade(Items):
    >>>     FRESHMAN = 1, 'FR', 0xE35314, 'Freshman'
    >>>     SOPHOMORE = 2, 'SO', 0xED15B4, 'Sophomore'
    >>>     JUNIOR = 3, 'JR', 0x9B3CED, 'Junior'
    >>>     SENIOR = 4, 'SR', 0xA0408E, 'Senior'
    >>>     GRADUATE = 5, 'GR', 0x518CCC, 'Graduate'
    >>>
    >>>     __properties__ = 'code', 'color', 'label'
    >>>
    >>>     @property
    >>>     def code(self) -> str:
    >>>         return self._code_
    >>>
    >>>     @property
    >>>     def color(self) -> int:
    >>>         return self._color_
    >>>
    >>>     @property
    >>>     def label(self) -> str:
    >>>         return self._label_
    >>>
    >>> Grade.SENIOR.name
    SENIOR

    >>> Grade.SENIOR.value
    4

    >>> Grade.SENIOR.code
    SR

    >>> hex(Grade.SENIOR.color)
    0xa0408e

    >>> Grade.SENIOR.label
    Senior

    >>> Grade.names
    ['FRESHMAN', 'SOPHOMORE', ...]

    >>> Grade.values
    [1, 2, 3, 4, 5]

    >>> Grade.codes
    ['FR', 'SO', 'JR', 'SR', 'GR']

    >>> Grade.colors
    [14897940, 15537588, ...]

    >>> Grade.labels
    ['Freshman', 'Sophomore', ...]
    """

    __properties__ = ()

    def __str__(self):
        return str(self.value)

    def __repr__(self):
        return f"{self.__class__.__qualname__}.{self._name_}"



"""
日期时间相关工具。
"""
from __future__ import annotations

__all__ = [
    'BearTimer',
    'is_leap',
    'get_last_monthday',
    'TimeFrame',
    'Datetime',
    'Timedelta',
    'daterange',
    'weekrange',
]

import logging
import sys
from datetime import datetime, timedelta, MINYEAR, MAXYEAR, date, time, tzinfo
from enum import IntEnum
from functools import wraps
from threading import Lock
from typing import Generator, Optional

logger = logging.getLogger('zeraora.datetime')


class BearTimer:
    __slots__ = '_label', '_marks', '_point'

    CONFIG = dict(
        version=1,
        formatters={
            'bear': dict(
                format='[%(asctime)s] [%(levelname)s] %(message)s',
            ),
            'bear_plus': dict(
                format='[%(asctime)s] [%(levelname)s] '
                       '[%(module)s.%(funcName)s:%(lineno)d] '
                       '%(message)s',
            ),
        },
        filters={},
        handlers={
            'Console': {
                'level': 'DEBUG',
                'class': 'logging.StreamHandler',
                'filters': [],
                'formatter': 'bear',
            },
        },
        loggers={
            'zeraora.datetime': dict(
                level='DEBUG',
                handlers=['Console'],
                propagate=False,
            ),
        },
    )

    def __init__(self, label: str = None, *_, **__):
        """
        熊牌秒表。对代码运行进行计时，并向名为 "zeraora.datetime" 的 Logger 发送 DEBUG 等级的日志。

        ----

        使用前，需要先启用日志输出： ::

            import logging.config
            from zeraora.datetime import BearTimer

            logging.config.dictConfig(BearTimer.CONFIG)
            bear = BearTimer()

        对于使用 Django 的项目可以改为在 settings.py 中进行如下设置： ::

            LOGGING = {
                'version': 1,
                'formatters': {...},
                'filters': {...},
                'handlers': {
                    'Console': {  # 确保有一个控制台输出
                        'level': 'DEBUG',
                        'class': 'logging.StreamHandler',
                    },
                    ...
                },
                'loggers': {
                    'zeraora.datetime': {  # 添加相应的日志记录器
                        'level': 'DEBUG',
                        'handlers': ['Console'],
                    },
                },
            }

        ----

        最简单是使用 ``with`` 语句包裹需要计时的部分：

        >>> with BearTimer() as bear:
        >>>     # 业务逻辑
        >>>     pass

        如需对一整个函数进行计时，可以作为装饰器使用：

        >>> @BearTimer()
        >>> def do_somthing(*args, **kwargs):
        >>>     # 业务逻辑
        >>>     pass

        带有多个装饰器时，计时器放哪里取决于你的计时范围：

        >>> from rest_framework.decorators import api_view
        >>> from zeraora.datetime import BearTimer
        >>>
        >>> @BearTimer()  # 从请求转发过来那一刻开始计时
        >>> @api_view(['GET'])
        >>> def query_status(request, *args, **kwargs):
        >>>     # 业务逻辑
        >>>     pass
        >>>
        >>> @api_view(['POST'])
        >>> @BearTimer()  # 从login()执行那一刻开始计时
        >>> def login(request, *args, **kwargs):
        >>>     # 业务逻辑
        >>>     pass

        此外还可以实例化一个对象。每个对象都是独立的计时器，互不影响。

        >>> bear = BearTimer()
        >>> bear.start()
        >>> # 业务逻辑
        >>> bear.stop()

        :param label: 计时器的标题，用以标明输出信息归属于哪个计时器。默认从打印消息时的上下文中获取。
        """
        try:
            # noinspection PyUnresolvedReferences,PyProtectedMember
            context = sys._getframe().f_back.f_code.co_name
        except AttributeError:
            context = ''
        self._label = context if not label and hasattr(sys, '_getframe') else str(label)
        self._marks: list[tuple[datetime, timedelta]] = []
        self._point: int = 0

    def __call__(self, func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            self._label = func.__name__
            self.start()
            returns = func(*args, **kwargs)
            self.stop()
            return returns

        return wrapper

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, traceback):
        self.stop()

    @property
    def records(self) -> list[tuple[datetime, timedelta]]:
        """
        每一次记录的时刻，以及距上一次记录的时间差。
        """
        return self._marks[:]

    def _log(self, total: float, delta: float, msg=''):
        logger.debug(f'[{self._label}] [{total:.9f} +{delta:.9f}]: {msg}')

    def start(self, msg='Starting...'):
        """
        清除之前的所有标记，并重新开始计时。

        :param msg: 要输出的消息。
        :return: 当前秒表。
        """
        self._marks = [(datetime.now(), timedelta())]
        self._log(0, 0, msg)
        return self

    def lap(self, msg='') -> tuple[datetime, datetime, datetime, timedelta, timedelta]:
        """
        记下当前的时刻。

        :param msg: 要输出的消息。
        :return: 开始时刻、上次记录时刻、当前时刻、距开始时间、距上次时间。
        """
        with Lock():
            if not self._marks:
                self.start('Automatically start...')

            head = self._marks[0][0]
            prev = self._marks[-1][0]
            curr = datetime.now()
            total = curr - head
            delta = curr - prev
            self._log(total.total_seconds(), delta.total_seconds(), msg)
            self._marks.append((curr, delta))
            return head, prev, curr, total, delta

    def stop(self, msg='Stopped.') -> datetime:
        """
        停止计时，但不清除所有标记。

        :param msg: 要输出的消息。
        :return: 当前时刻。
        """
        _, _, curr, _, _ = self.lap(msg)
        return curr


def is_leap(year: int) -> bool:
    """
    判断一个年份是否为闰年。
    """
    return year % 4 == 0 and (year % 100 != 0 or year % 400 == 0)


def get_last_monthday(year: int, month: int) -> int:
    """
    计算某年某月的最后一天是哪一天。
    """
    if month == 2 and is_leap(year):
        return 29
    return Datetime.DAYS_IN_MONTH[month]


class TimeFrame(IntEnum):
    YEAR = 0
    MONTH = 1
    DAY = 2
    HOUR = 3
    MINUTE = 4
    SECOND = 5
    MICROSECOND = 6


class Datetime(datetime):
    """
    增强型日期时间对象。
    """

    DAYS_IN_MONTH = (-1, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31)

    # ---- 构造器 ----

    @classmethod
    def fromcalendar(cls, year: int, week: int, day: int, sunday_first=False) -> Datetime:
        """
        将某一年某一周的星期几转换为一个具体的日期。

        - ``sunday_first=False`` 时一年内第一个周一之前的日子都属于第 ``0`` 周。
        - ``sunday_first=True`` 时一年内第一个周日之前的日子都属于第 ``0`` 周。

        :param year: 具体年份。比如 2012、2023 等。
        :param week: 第几周。从 ``0`` 开始。
        :param day: 星期几。``0`` 表示周日、``1`` 表示周一，以此类推。
        :param sunday_first: 是否将周日作为一周的开始。
        :return: 一个具体的日期。
        """
        return cls.strptime(
            f'{year:04d}-{week:02d}-{day:1d}',
            '%Y-%U-%w' if sunday_first else '%Y-%W-%w',
        )

    @classmethod
    def fromdatetime(cls, _datetime: datetime) -> Datetime:
        """
        用 :class:`datetime` 构造对象。
        """
        return cls(
            _datetime.year, _datetime.month, _datetime.day,
            _datetime.hour, _datetime.minute, _datetime.second,
            _datetime.microsecond, _datetime.tzinfo, fold=_datetime.fold,
        )

    of = fromdatetime

    # ---- 转换器 ----

    def replace(
            self,
            year: Optional[int] = None,
            month: Optional[int] = None,
            day: Optional[int] = None,
            hour: Optional[int] = None,
            minute: Optional[int] = None,
            second: Optional[int] = None,
            microsecond: Optional[int] = None,
            _tzinfo: Optional[tzinfo] = ...,
            *,
            fold: Optional[int] = None,
            standard: bool = False,
    ) -> Datetime | datetime:
        """
        修改部分字段并返回一个新的日期时间。

        如果 ``standard=True`` 则固定返回 :class:`datetime` 对象。
        """
        return (datetime if standard else type(self))(
            year or self.year,
            month or self.month,
            day or self.day,
            hour or self.hour,
            minute or self.minute,
            second or self.second,
            microsecond or self.microsecond,
            self.tzinfo if _tzinfo is ... else _tzinfo,
            fold=fold or self.fold,
        )

    @classmethod
    def combine(
            cls,
            a: date | datetime,
            b: time | datetime | timedelta,
            _tzinfo: Optional[tzinfo] = ...
    ) -> Datetime:
        """
        将 a 的日期部分和 b 的时间部分组合为一个新的日期时间。

        :param a: 支持 :class:`data` 或 :class:`datetime` 及其子类的对象。
        :param b: 支持 :class:`time`、:class:`datetime` 或 :class:`timedelta` 及其子类的对象。
        :param _tzinfo: 时区。默认使用 b 的时区设置，如不需要可设置为 ``None`` 。
        :return: 组合得到的日期时间。
        """
        if _tzinfo is ...:
            _tzinfo = b.tzinfo
        if not isinstance(a, (date, datetime)):
            raise TypeError("argument 'a' must be a date or datetime instance")
        if not isinstance(b, (time, datetime, timedelta)):
            raise TypeError("argument 'b' must be a time, datetime, or timedelta instance")
        if isinstance(b, timedelta):
            b = Timedelta.of(b).base()

        return cls(
            a.year, a.month, a.day,
            b.hour, b.minute, b.second, b.microsecond,
            _tzinfo, fold=b.fold,
        )

    def empty(
            self,
            keeping_level: TimeFrame = TimeFrame.DAY,
            year: Optional[int] = None,
            month: Optional[int] = None,
            day: Optional[int] = None,
            hour: Optional[int] = None,
            minute: Optional[int] = None,
            second: Optional[int] = None,
            microsecond: Optional[int] = None,
    ) -> Datetime:
        """
        将某个层级以下（不含）的部分归零。

        >>> now = Datetime.now()
        >>> for frame in TimeFrame:
        >>>     print(
        >>>         now.empty(frame), frame, sep='\\t'
        >>>     )
        2024-01-01 00:00:00         TimeFrame.YEAR
        2024-09-01 00:00:00         TimeFrame.MONTH
        2024-09-17 00:00:00         TimeFrame.DAY
        2024-09-17 11:00:00         TimeFrame.HOUR
        2024-09-17 11:22:00         TimeFrame.MINUTE
        2024-09-17 11:22:33         TimeFrame.SECOND
        2024-09-17 11:22:33.245678  TimeFrame.MICROSECOND

        可以将时间的某个部分固定为指定值，比如将小时固定为凌晨四点：

        >>> for frame in TimeFrame:
        >>>     print(
        >>>         now.empty(frame, hour=4), frame, sep='\\t'
        >>>     )
        2024-01-01 04:00:00         TimeFrame.YEAR
        2024-09-01 04:00:00         TimeFrame.MONTH
        2024-09-17 04:00:00         TimeFrame.DAY
        2024-09-17 04:00:00         TimeFrame.HOUR
        2024-09-17 04:22:00         TimeFrame.MINUTE
        2024-09-17 04:22:33         TimeFrame.SECOND
        2024-09-17 04:22:33.245678  TimeFrame.MICROSECOND
        """
        index = keeping_level + 1
        origin = (
            year or self.year,
            month or self.month,
            day or self.day,
            hour or self.hour,
            minute or self.minute,
            second or self.second,
            microsecond or self.microsecond,
        )
        specified = (
            year or MINYEAR,
            month or 1,
            day or 1,
            hour or 0,
            minute or 0,
            second or 0,
            microsecond or 0,
        )
        return type(self)(
            *(origin[:index] + specified[index:]),
            tzinfo=self.tzinfo, fold=self.fold,
        )

    def fill(
            self,
            keeping_level: TimeFrame = TimeFrame.DAY,
            year: Optional[int] = None,
            month: Optional[int] = None,
            day: Optional[int] = None,
            hour: Optional[int] = None,
            minute: Optional[int] = None,
            second: Optional[int] = None,
            microsecond: Optional[int] = None,
    ) -> Datetime:
        """
        将某个层级以下（不含）的部分填满。

        >>> now = Datetime.now()
        >>> for frame in TimeFrame:
        >>>     print(
        >>>         now.fill(frame), frame, sep='\\t'
        >>>     )
        2024-12-31 23:59:59.999999    TimeFrame.YEAR
        2024-09-30 23:59:59.999999    TimeFrame.MONTH
        2024-09-09 23:59:59.999999    TimeFrame.DAY
        2024-09-09 11:59:59.999999    TimeFrame.HOUR
        2024-09-09 11:22:59.999999    TimeFrame.MINUTE
        2024-09-09 11:22:33.999999    TimeFrame.SECOND
        2024-09-09 11:22:33.245678    TimeFrame.MICROSECOND

        可以将时间的某个部分固定为给定值，比如将小时固定为凌晨四点：

        >>> for frame in TimeFrame:
        >>>     print(
        >>>         now.fill(frame), frame, sep='\\t'
        >>>     )
        2024-12-31 04:59:59.999999    TimeFrame.YEAR
        2024-09-30 04:59:59.999999    TimeFrame.MONTH
        2024-09-09 04:59:59.999999    TimeFrame.DAY
        2024-09-09 04:59:59.999999    TimeFrame.HOUR
        2024-09-09 04:22:59.999999    TimeFrame.MINUTE
        2024-09-09 04:22:33.999999    TimeFrame.SECOND
        2024-09-09 04:22:33.245678    TimeFrame.MICROSECOND
        """
        index = keeping_level + 1
        origin = [
            year or self.year,
            month or self.month,
            day or self.day,
            hour or self.hour,
            minute or self.minute,
            second or self.second,
            microsecond or self.microsecond,
        ]
        specified = [
            year or MAXYEAR,
            month or 12,
            day or 31,
            hour or 23,
            minute or 59,
            second or 59,
            microsecond or 999999,
        ]
        result = origin[:index] + specified[index:]
        result[2] = min(result[2], get_last_monthday(result[0], result[1]))
        return type(self)(*result, tzinfo=self.tzinfo, fold=self.fold)

    def calendar(self, sunday_first=False) -> tuple[int, int, int]:
        """
        计算当前日期是哪一年哪一周的周几。

        - ``sunday_first=False`` 时一年内第一个周一之前的日子都属于第 ``0`` 周。
        - ``sunday_first=True`` 时一年内第一个周日之前的日子都属于第 ``0`` 周。

        :param sunday_first: 是否将周日作为一周的开始。
        :return: 一个三元组，分别表示哪一、哪一周、周几。
        """
        return tuple(map(int, self.strftime('%Y-%U-%w' if sunday_first else '%Y-%W-%w').split('-')))

    # ---- 判断 ----

    def is_leap(self) -> bool:
        """
        当年是否为闰年。
        """
        return self.year % 4 == 0 and (self.year % 100 != 0 or self.year % 400 == 0)

    # ---- 属性 ----

    @property
    def last_monthday(self) -> int:
        """
        当月的最后一天是哪一天。
        """
        if self.month == 2 and self.is_leap():
            return 29
        return self.DAYS_IN_MONTH[self.month]


class Timedelta(timedelta):
    """
    增强型时间差对象。
    """

    # ---- 构造器 ----

    @classmethod
    def fromtime(cls, _time: time) -> Timedelta:
        """
        用 :class:`time` 构造时间差对象。

        换句话说就是 fromtime(_time) = _time - 00:00:00
        """
        return cls(
            seconds=_time.hour * 3600 + _time.minute * 60 + _time.second,
            microseconds=_time.microsecond,
        )

    @classmethod
    def fromtimedelta(cls, delta: timedelta) -> Timedelta:
        """
        用 :class:`timedelta` 构造时间差对象。
        """
        return cls(days=delta.days, seconds=delta.seconds, microseconds=delta.microseconds)

    of = fromtimedelta

    # ---- 转换 ----

    def replace(
            self,
            days: Optional[int] = None,
            seconds: Optional[int] = None,
            microseconds: Optional[int] = None,
            standard: bool = False,
    ) -> Timedelta | timedelta:
        """
        修改部分字段并返回一个新的时间差对象。

        如果 ``standard=True`` 则固定返回 :class:`timedelta` 对象。
        """
        return (timedelta if standard else type(self))(
            days or self.days,
            seconds or self.seconds,
            microseconds or self.microseconds,
        )

    def base(self, _time: Optional[time] = None) -> time:
        """
        基于某个时间叠加得到另一个时间。

        **该方法可能会在将在被废弃**！
        """
        if _time is None:
            s = self.seconds
            ms = self.microseconds
        else:
            ms = _time.microsecond + self.microseconds
            s = ms // 10 ** 6 + self.seconds
            ms = ms % 10 ** 6
        return time(
            hour=s // 3600 % 24,
            minute=s % 3600 // 60,
            second=s % 60,
            microsecond=ms,
        )


def daterange(start: date, stop: date, step: int = 1, closed=False) -> Generator[date, None, None]:
    """
    在一个日期范围内，按指定的步长迭代生成 :class:`date` 对象。

    :param start: 开始日期。
    :param stop: 结束日期。
    :param step: 步长。提供负数时，请确保开始日期晚于（即大于）结束日期。如果为 ``0`` 会引发 :class:`datetime` 。
    :param closed: 结束日期是否可以到达。
    :return: 生成 :class:`date` 对象的生成器。
    """
    if closed:
        ordinals = range(start.toordinal(), stop.toordinal() + step, step)
    else:
        ordinals = range(start.toordinal(), stop.toordinal(), step)

    return (date.fromordinal(o) for o in ordinals)


def weekrange(year: int, week: int, sunday_first=False) -> Generator[date, None, None]:
    """
    枚举生成某一周的所有日期。

    :param year: 具体年份。比如 2012、2023 等。
    :param week: 一年中的第几周。从 ``0`` 开始。
    :param sunday_first: 是否以周日为一周的开始。
    :return: 生成 :class:`date` 对象的生成器。
    """
    start = datetime.strptime(
        f'{year:04d}-{week:02d}-{0 if sunday_first else 1}',
        '%Y-%U-%w' if sunday_first else '%Y-%W-%w',
    ).date()
    return (start + timedelta(days=i) for i in range(7))



"""
数学相关工具和常量。
"""
from __future__ import annotations

__all__ = [
    'NaN',
    'Decimal',
    'remove_exponent',
    'absolute',
    'bitstream',
    'digitstream',
]

from decimal import Decimal as StandardDecimal
from typing import Generator

NaN = float('NaN')
"""二进制小数型 ``NaN`` ，即 Not a Number（非数值）。"""


class Decimal(StandardDecimal):
    # 同名是为了方便无感替换

    NAN = StandardDecimal('NaN')
    """一个非数值的值（Not a Number）。"""

    ZERO = StandardDecimal(0)
    """整数 ``0`` 。"""

    ONE = StandardDecimal(1)
    """整数 ``1`` 。"""

    PI = StandardDecimal('3.141592653589793238462643383279')
    """包含 30 位小数的圆周率。"""

    E = StandardDecimal('2.718281828459045')
    """包含 15 位小数的自然常数。"""

    def remove_exponent(self) -> Decimal:
        """
        去除十进制小数的尾导零。

        摘录自 `Decimal 常见问题 <https://docs.python.org/zh-cn/3/library/decimal.html#decimal-faq>`_ 。
        """
        return self.quantize(self.ONE) if self == self.to_integral() else self.normalize()


def remove_exponent(d: Decimal) -> Decimal:
    """
    去除十进制小数的尾导零。

    摘录自 `Decimal 常见问题 <https://docs.python.org/zh-cn/3/library/decimal.html#decimal-faq>`_ 。
    """
    return d.quantize(Decimal.ONE) if d == d.to_integral() else d.normalize()


def absolute(n: int | float | Decimal) -> tuple[bool, int | float | Decimal]:
    """
    返回一个实数的符号及其绝对值。``True`` 表示正数，``False`` 表示负数。
    """
    return n >= 0, abs(n)


def bitstream(integer: int) -> Generator[int, None, None]:
    """
    获取一个整数的所有比特位。

    >>> list(bitstream(67))
    [1, 2, 64]

    >>> list(bitstream(-43))
    [-1, -2, -8, -32]

    >>> list(bitstream(0))
    []
    """
    sign, positive = absolute(integer)
    if not isinstance(positive, int):
        yield from []
        return
    for power in range(0, positive.bit_length()):
        bit = 1 << power
        if positive & bit:
            yield bit if sign else -bit


def digitstream(integer: int, base: int) -> Generator[int, None, None]:
    """
    获取一个非负整数在 *base* 进制下的各位数码，**以逆序生成** 。

    生成器无法直接反转顺序，建议调用者处理好生成结果后再取出对象、进行反转。

    >>> digits = digitstream(1008612, 16)
    >>> mapper = '0123456789abcdef'.__getitem__
    >>> ''.join(map(mapper, digits))[::-1]
    'f63e4'
    >>> hex(1008612)
    '0xf63e4'

    :param integer: 十进制整数。参数会被取绝对值，需要提前保留整数的符号。
    :param base: 需要转换为什么进位制。参数不能小于 ``2`` 。
    :return: 一个迭代器，每次迭代会 “从右到左” 输出结果的一位的十进制表示。
    """
    if base < 2:
        yield from []
        return
    integer = abs(integer)
    while integer >= base:
        yield integer % base
        integer //= base
    yield integer



"""
二进制相关。
"""

__all__ = [
    'randbytes',
]

from random import getrandbits


def randbytes(n: int) -> bytes:
    """
    生成 n 个随机字节。

    此函数用于在 Python 3.9 以前提供标准库中 random.randbytes(n) 的等效能力。
    """
    if n < 1:
        return b''
    return getrandbits(n * 8).to_bytes(n, 'little')



"""
对 `requests <https://pypi.org/project/requests/>`_ 的扩展和增强。
"""

__all__ = [
    'HTTPBearerAuth',
]

from requests.auth import AuthBase


class HTTPBearerAuth(AuthBase):
    """
    为 ``requests.post()`` 等方法的 *auth* 参数提供对象。
    """

    def __init__(self, token):
        self.token = token

    def __call__(self, r):
        r.headers['Authorization'] = f'Bearer {self.token}'
        return r

    def __eq__(self, other):
        return self.token == getattr(other, 'token', None)

    def __ne__(self, other):
        return self.token != getattr(other, 'token', None)



"""
Zeraora

A personal utility package, with long time supports.
"""

VERSION = (0, 4, 0, 'alpha', 1)

# https://packaging.python.org/en/latest/specifications/version-specifiers/
__version__ = '0.4.0a1'
